import json
from typing import Callable, Optional

from fabric.tasks import task

from xefab.utils import console

from .shell import get_system, which

CONDA_LINKS = {
    "linux": "https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh",
    "darwin": "https://repo.anaconda.com/miniconda/Miniconda3-latest-MacOSX-x86_64.sh",
    "windows": "https://repo.anaconda.com/miniconda/Miniconda3-latest-Windows-x86_64.exe",
}


def ensure_dependency(dep, local: bool = False, installer: Optional[Callable] = None):
    @task
    def wrapper(c, *args, **kwargs):
        if which(c, dep, local=local, hide=True):
            return
        if installer is not None:
            installer(c, *args, **kwargs)
        else:
            raise RuntimeError(f"Dependency {dep} not found.")

    return wrapper


@task
def go(c, force: bool = False):
    """Install Go."""

    if which(c, "go", hide=True) and not force:
        console.print("Go already installed on system.")
        return

    system = get_system(c, hide=True)
    if get_system(c, hide=True) in ("Linux", "Darwin"):
        c.run(
            "wget -q -O - https://raw.githubusercontent.com/canha/golang-tools-install-script/master/goinstall.sh | bash"
        )
        console.print(
            "Go installed. Please add `export PATH=$PATH:/usr/local/go/bin` to your .bashrc/.profile file"
        )
    else:
        raise RuntimeError(f"{system} currently not supported by this task.")


@task
def gopass(c, force: bool = False):
    """Install gopass."""

    if which(c, "gopass", hide=True) and not force:
        console.print("Gopass already installed on system.")
        return
    c.run("go install github.com/gopasspw/gopass@latest")
    console.print("Done")


@task
def chezmoi(c, force: bool = False):
    """Install chezmoi."""

    if which(c, "chezmoi", hide=True) and not force:
        console.print("Chezmoi already installed on system.")
        return
    c.run('sh -c "$(curl -fsLS get.chezmoi.io)"')
    console.print("Done.")


@task(aliases=["gh"])
def github_cli(c, force: bool = False):
    """Install the Github CLI."""

    console.print("Checking for existing installation.")
    have_gh = which(c, "gh", hide=True)

    if have_gh and not force:
        console.print("Github CLI already installed on system.")
        return

    console.print("Attempting to install Github CLI.")

    if which(c, "conda", hide=True):
        console.print("Installing Github CLI via conda.")
        c.run("conda install gh --channel conda-forge")
        return

    console.print("Finding latest release.")

    command = f"gh api /repos/cli/cli/releases/latest"
    if hasattr(c, "local"):
        result = c.local(command, hide=True, warn=True)
    else:
        result = c.run(command, hide=True, warn=True)
    assert result.ok, "Failed to get latest release."
    latest_release = json.loads(result.stdout)
    console.print(f"Installing release: {latest_release['tag_name']}")
    assets = latest_release["assets"]

    system = get_system(c, hide=True)

    if system == "Linux":
        asset = [a for a in assets if a["name"].endswith("linux_amd64.tar.gz")][0]
    elif system == "Darwin":
        asset = [a for a in assets if a["name"].endswith("darwin_amd64.tar.gz")][0]
    else:
        raise RuntimeError(f"{system} currently not supported by this task.")

    console.print(
        f"Downloading asset: {asset['name']} from {asset['browser_download_url']}"
    )
    c.run(f"wget -q {asset['browser_download_url']}")

    c.run(f"tar -xzf {asset['name']}")

    c.run(
        f"mv {asset['name'].replace('.tar.gz', '')}/bin/gh /home/{c.user}/.local/bin/gh"
    )
    console.print(
        f"gh installed. Please ensure `/home/{c.user}/.local/bin` is in your path."
    )
    console.print("Done.")


@task(
    aliases=["gpg"],
)
def gnupg(c, force: bool = False):
    """Install GnuPG."""

    if which(c, "gpg", hide=True) and not force:
        console.print("GnuPG already installed on system.")
        return
    scripts = {
        "install_gpg_all.sh": "https://raw.githubusercontent.com/rnpgp/gpg-build-scripts/master/install_gpg_all.sh",
        "install_gpg_component.sh": "https://raw.githubusercontent.com/rnpgp/gpg-build-scripts/master/install_gpg_component.sh",
    }

    for name, script in scripts.items():
        c.run(f"wget -q {script}")
        c.run(f"chmod +x {name}")
    c.run("./install_gpg_all.sh")
    console.print("Done.")


@task(
    aliases=["conda"],
)
def miniconda(c, download_only: bool = False):
    """Install Miniconda."""

    system = get_system(c, hide=True).lower()
    if system not in CONDA_LINKS:
        raise RuntimeError(f"{system} currently not supported by this task.")
    link = CONDA_LINKS[system]
    with console.status(f"Downloading miniconda installer from {link}"):
        result = c.run(f"wget -q {link}", hide=True, warn=True)
    if result.failed:
        raise RuntimeError(f"Failed to download miniconda installer.")

    if download_only:
        console.print(f"Miniconda installer downloaded.")
        return

    if system == "windows":
        command = """start /wait "" Miniconda3-latest-Windows-x86_64.exe /InstallationType=JustMe /RegisterPython=0 /S /D=%UserProfile%\Miniconda3
        """
        c.run(command)
        return
    elif system == "darwin":
        c.run(f"chmod +x Miniconda3-latest-MacOSX-x86_64.sh")
        c.run(f"./Miniconda3-latest-MacOSX-x86_64.sh")
    elif system == "linux":
        c.run(f"chmod +x Miniconda3-latest-Linux-x86_64.sh")
        c.run(f"./Miniconda3-latest-Linux-x86_64.sh  -b -p $HOME/miniconda")
    console.print("Done.")



from invoke.collection import Collection
from invoke.config import merge_dicts
from invoke.tasks import Task
from invoke.util import debug
from paramiko.config import SSHConfig

from xefab.utils import console

from .entrypoints import get_entry_points


class XefabCollection(Collection):
    """Collection that assigns host name
    as default for all its tasks.
    """

    entrypoint_prefix = "xefab"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def task_with_config(self, name):
        # fetch the task object
        task, conf = super().task_with_config(name)

        # set default hosts for task
        if "hostnames" in self._configuration:
            task.hosts = [self.name]
        return task, conf

    def load_objects_from_entry_points(self, group="tasks"):
        """Load tasks/collections from entry points."""

        if not (group and isinstance(group, str)):
            raise ValueError(
                f"Entrypoint groups must be non-empty strings. got {group}"
            )

        if not group.startswith(self.entrypoint_prefix):
            group = ".".join([self.entrypoint_prefix, group])

        for ep in get_entry_points(group):
            try:
                obj = ep.load()
                if isinstance(obj, type):
                    obj = obj()
                self._add_object(obj, name=ep.name)
                obj = self.collections.get(ep.name, None)
                if isinstance(obj, XefabCollection):
                    grp = ".".join([group, ep.name])
                    obj.load_objects_from_entry_points(grp)
            except TypeError as e:
                debug(
                    f"xefab.{self.name}: Error loading tasks from {ep.name} due to wrong type. "
                    f"details: {e}"
                )
                continue
            except ImportError as e:
                debug(
                    f"xefab.{self.name}: Error loading tasks from {ep.name} due to import error."
                    f"details: {e}"
                )
                continue
            except Exception as e:
                debug(
                    f"xefab.{self.name}: Error loading tasks from {ep.name} due to unknwon error."
                    f"details: {e}"
                )
                continue



import contextlib
import errno
import os
import re
import socket
from collections import defaultdict
from inspect import Parameter
from typing import Iterable, Optional, Tuple, Union

import fsspec
import pandas as pd
from decopatch import DECORATED, function_decorator
from fabric.connection import Connection
from invoke.context import Context
from invoke.util import enable_logging
from makefun import wraps
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.progress import (Progress, ProgressColumn, SpinnerColumn, Task,
                           TextColumn)
from rich.table import Table
from rich.theme import Theme

custom_theme = Theme({"info": "dim cyan", "warning": "magenta", "danger": "bold red"})

console = Console(theme=custom_theme)


SHELL_PROFILE_FILES = {
    "sh": ["~/.profile"],
    "bash": ["~/.profile", "~/.bash_profile"],
    "zsh": ["~/.profile", "~/.zprofile"],
}


if os.environ.get("XEFAB_DEBUG") in ("1", "true", "True"):
    enable_logging()


def nested_dict():
    """Create a nested defaultdict."""
    return defaultdict(nested_dict)


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


@function_decorator
def try_local(exception=Exception, f=DECORATED):
    """Try to run a task locally, if it fails, run it remotely.
    Can be used as a decorator or as a function.
    if the first argument after the function is a Connection
    the function is called immediately, otherwise a wrapper is returned.
    """
    extra = Parameter(
        "force_remote",
        kind=Parameter.POSITIONAL_OR_KEYWORD,
        annotation=bool,
        default=False,
    )

    @wraps(f, append_args=extra)
    def wrapper(c, *args, force_remote=False, **kwargs):
        if isinstance(c, Connection) and not force_remote:
            try:
                inv_ctx = Context(config=c.config)
                return f(inv_ctx, *args, **kwargs)
            except exception as e:
                console.print(f"Failed to run locally: {e}")
                console.print("Running remotely instead.")
        return f(c, *args, **kwargs)

    return wrapper


def filesystem(c: Union[Connection, Context]):
    """Get a fsspec filesystem object from a fabric Connection/Invoke Context object."""
    if isinstance(c, Connection):
        return fsspec.filesystem(
            "sftp", host=c.host, username=c.user, port=c.port, **c.connect_kwargs
        )
    else:
        return fsspec.filesystem("file")


def get_open_port(start=5000, end=None, bind_address="", *socket_args, **socket_kwargs):
    if start < 1024:
        start = 1024

    if end is None:
        end = start + 10000
    port = start
    while port < end:
        try:
            with contextlib.closing(
                socket.socket(*socket_args, **socket_kwargs)
            ) as my_socket:
                my_socket.bind((bind_address, port))
                my_socket.listen(1)
                this_port = my_socket.getsockname()[1]
                return this_port
        except socket.error as error:
            if not error.errno == errno.EADDRINUSE:
                raise
        port += 1
    raise Exception("Could not find open port")


def df_to_table(
    pandas_dataframe: pd.DataFrame,
    rich_table: Table = None,
    show_index: bool = True,
    index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.
    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        rich_table (Table): A rich Table that should be populated by the DataFrame values.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: The rich Table instance passed, populated with the DataFrame values."""
    if rich_table is None:
        rich_table = Table(show_header=True, header_style="bold magenta")
    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table


class SuccessSpinnerColumn(SpinnerColumn):
    """A spinner column that shows a checkmark when the task
    is completed successfully or a x if not.
    """

    def render(self, task: Task) -> RenderableType:
        if task.finished:
            if task.fields.get("exception", None) is None:
                return "[bold green]✓[/bold green]"
            else:
                return "[bold red]✗[/bold red]"
        else:
            return super().render(task)


class ProgressContext(Progress):
    """A context manager for rich.progress.Progress.
    Allows entering a task context and task will be completed when exiting
    """

    def __init__(self, *args, **kwargs):
        self._live_display = None
        super().__init__(*args, **kwargs)

    @classmethod
    def get_default_columns(cls) -> Tuple[ProgressColumn, ...]:
        return (
            SuccessSpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            SpinnerColumn(spinner_name="simpleDots", finished_text=""),
        )

    def get_renderables(self) -> Iterable[RenderableType]:
        yield from super().get_renderables()
        if self._live_display is not None:
            yield self._live_display

    @contextlib.contextmanager
    def enter_task(
        self,
        description,
        total=1,
        finished_description=None,
        exception_description="Task failed to excecute.",
        warn=False,
        hide=False,
    ):
        """Start and end a task in a progress bar."""
        task = self.add_task(description, total=total)
        exception = None
        try:
            yield task
        except Exception as e:
            finished_description = exception_description.format(exception=e)
            exception = e
            self.update(task_id=task, exception=True)
        finally:
            description = finished_description or self.tasks[task].description
            self.update(task_id=task, completed=total, description=description)
            if exception is not None and not hide:
                self.console.print_exception(show_locals=True)
            if exception is not None and not warn:
                exit(1)

    def live_display(self, renderable):
        """Display a renderable in a panel below the progress bar."""
        with self._lock:
            self._live_display = renderable
        self.refresh()


def tail(text, n=10):
    """Return the last n lines of a text string."""
    return "\n".join(text.splitlines()[-n:])



import json
import time
import uuid
from io import BytesIO, StringIO
from typing import List

from fabric.connection import Connection
from fabric.tasks import task
from rich.panel import Panel

from xefab.tasks.shell import exists, is_file
from xefab.tasks.squeue import parse_squeue_output
from xefab.utils import ProgressContext, console, tail

SLURM_INSTRUCTIONS = {
    "partition": "partition to submit the job to.",
    "qos": "quality of service to submit the job to.",
    "time": "number of hours to run the job for, can be a `hrs:mins:secs` string or a number.",
    "mem_per_cpu": "memory per cpu.",
    "cpus_per_task": "number of cpus per task.",
    "job_name": "name of the job.",
    "job": "jobscript to run.",
    "output": "where to save the output of the job.",
    "error": "where to save the error of the job.",
    "account": "account to submit the job to.",
    "chdir": "change directory to this path before running the job.",
    "mail_type": "when to send an email about the job.",
    "mail_user": "email address to send the email to.",
    "array": "array of jobs to run.",
    "dependency": "job dependency.",
    "exclude": "nodes to exclude.",
    "gres": "generic resources to use.",
    "hint": "hint to the scheduler.",
    "kill_on_invalid_dep": "kill the job if the dependency is invalid.",
    "nodelist": "nodes to use.",
    "ntasks": "number of tasks to run.",
    "ntasks_per_node": "number of tasks per node.",
    "ntasks_per_core": "number of tasks per core.",
    "nodes": "number of nodes to use.",
    "overcommit": "overcommit resources.",
    "requeue": "requeue the job if it fails.",
    "reservation": "reservation to use.",
    "threads_per_core": "threads per core.",
    "wait": "wait for the job to finish.",
    "comment": "comment to add to the job.",
    "constraint": "constraint to use.",
    "get_user_env": "get the user environment.",
}

SINGULARITY_ARGUMENTS = {
    "bind": "binds a directory to the container.",
}


SBATCH_TEMPLATE = """#!/bin/bash
{slurm_instructions}

{env_settings}

unset X509_CERT_DIR CUTAX_LOCATION

module load singularity

{command}

echo {done_message}
echo "Finished at $(date)"
"""


def generate_slurm_instructions(**kwargs):
    """Generates the singularity command for the sbatch script"""
    instructions = []
    for key, value in kwargs.items():
        if key in SLURM_INSTRUCTIONS:
            key = key.replace("_", "-")
            if key == "time":
                if not ":" in value:
                    value = f"{value}:00:00"
                if isinstance(value, int):
                    value = f"{value:02d}:00:00"
                elif isinstance(value, float):
                    value = f"{int(value):02d}:{int(value * 60 % 60):02d}:{int(value * 60 % 60 * 60 % 60):02d}"
                instructions.append(f"#SBATCH --time={value}")
            else:
                instructions.append(f"#SBATCH --{key}={value}")

    return "\n".join(instructions) + "\n"


def generate_singularity_instructions(command, image_path, **kwargs):
    """Generates the instructions for the singularity exec"""
    instructions = ["singularity exec"]
    for key, value in kwargs.items():
        if key not in SINGULARITY_ARGUMENTS:
            continue
        if key == "bind":
            if isinstance(value, str):
                value = value.split(",")
            bind_str = " ".join([f"--bind {b}" for b in value])
            instructions.append(bind_str)
        elif isinstance(value, bool):
            instructions.append(f"--{key}")
        else:
            instructions.append(f"--{key}={value}")

    instructions.append(image_path)
    instructions.append(command)

    return " ".join(instructions)


def generate_env_instructions(**env):
    """Generates the environment variables export commands for the job"""
    if not env:
        return ""
    instructions = []
    for key, value in env.items():
        instructions.append(f"export {key}={value}")

    return "\n".join(instructions) + "\n\n"


SINGULARITY_DIR = "/project2/lgrandi/xenonnt/singularity-images"


@task(
    help={
        "script": "the command/script to execute within the job.",
        "partition": "partition to submit the job to.",
        "qos": "qos to submit the job to.",
        "account": "account to submit the job to.",
        "job_name": "how to name this job.",
        "dry_run": "Just print the job file.",
        "mem_per_cpu": "mb requested for job.",
        "container": "name of the container to activate",
        "bind": "which paths to add to the container",
        "cpus_per_task": "cpus requested for job",
        "hours": "max hours of a job",
        "container_dir": "where to find the container",
        "env_dict": "environment variables to set before running the job",
        "workdir": "working directory for the job",
        "output": "where to save the stdout stream of the job",
        "error": "where to save the stderr stream of the job",
        "timeout": "how long to wait for the job to start before exiting (seconds)",
        "extra_instructions": "extra slurm instructions to add to the batch file",
    }
)
def sbatch(
    c: Connection,
    script: str,
    *,
    partition: str = "xenon1t",
    qos: str = "xenon1t",
    account: str = "pi-lgrandi",
    job_name: str = None,
    dry_run: bool = False,
    mem_per_cpu: int = 1000,
    container="xenonnt-development.simg",
    bind: list = None,
    cpus_per_task: int = 1,
    hours: float = 12,
    container_dir: str = SINGULARITY_DIR,
    env_dict: str = None,
    workdir: str = None,
    output: str = None,
    error: str = None,
    extra_instructions: str = None,
    timeout: int = 120,
):
    """
    Create and submit a job to SLURM job queue on the remote host.
    """

    job_id = str(uuid.uuid4()).replace("-", "")[:8]

    if job_name is None:
        job_name = f"xefab_job_{job_id}"

    if partition == "kicp":
        qos = "xenon1t-kicp"
    else:
        qos = partition

    # parse and check arguments

    if bind is None:
        bind = ["/dali", "/project2", "/scratch", "/cvmfs"]
    if len(bind) == 1:
        bind = bind[0].split(",")
    bind = [b.strip() for b in bind]

    if env_dict is None:
        env_dict = {}
    else:
        env_dict = json.loads(env_dict)

    if extra_instructions is None:
        extra_instructions = {}
    else:
        extra_instructions = json.loads(extra_instructions)

    with ProgressContext() as progress:
        if workdir is None:
            with progress.enter_task("Checking for $SCRATCH folder") as task:
                result = c.run("echo $SCRATCH", hide=True, warn=True)
            if result.ok and result.stdout:
                SCRATCH = result.stdout.strip()
            else:
                SCRATCH = f"/scratch/midway2/{c.user}"

            workdir = f"{SCRATCH}/xefab_jobs/{job_name}"

        if output is None:
            output = f"{workdir}/{job_name}.out"
        if error is None:
            error = f"{workdir}/{job_name}.err"

        if script.endswith(".py"):
            remote_script_path = f"{workdir}/{job_name}.py"
            command = f"python {remote_script_path}"
        else:
            remote_script_path = f"{workdir}/{job_name}.sh"
            command = remote_script_path

        sbatch_path = f"{workdir}/{job_name}.sbatch"

        with progress.enter_task(
            "Testing remote connection and workdir existince"
        ) as task:
            if not exists(c, workdir, hide=True):
                progress.update(
                    task, description=f"Creating workdir {workdir} on {c.host}"
                )
                c.run(f"mkdir -p {workdir}", hide=True)

        if is_file(c, script, hide=True, local=True):
            with progress.enter_task(f"Copying script to {c.original_host}"):
                c.put(script, remote=remote_script_path)

        elif is_file(c, script, hide=True):
            with progress.enter_task(f"Copying script to working directory"):
                c.run(f"cp {script} {remote_script_path}", hide=True)
        else:
            script_fd = StringIO("#!/bin/bash\n" + script)
            with progress.enter_task(f"Creating script at on {c.original_host}"):
                c.put(script_fd, remote=remote_script_path)

        if remote_script_path.endswith("sh"):
            with progress.enter_task(f"Making script executable"):
                c.run(f"chmod +x {remote_script_path}", hide=True)

        with progress.enter_task(f"Creating sbatch file"):
            slurm_instructions = generate_slurm_instructions(
                job_name=job_name,
                partition=partition,
                qos=qos,
                account=account,
                mem_per_cpu=mem_per_cpu,
                cpus_per_task=cpus_per_task,
                hours=hours,
                output=output,
                error=error,
                chdir=workdir,
                **extra_instructions,
            )

            env_instructions = generate_env_instructions(**env_dict)

            if container:
                image_path = f"{container_dir.rstrip('/')}/{container}"
                command = generate_singularity_instructions(
                    command, image_path, bind=bind
                )

            done_message = f"Job {job_name} done."

            sbatch_content = SBATCH_TEMPLATE.format(
                slurm_instructions=slurm_instructions,
                env_settings=env_instructions,
                command=command,
                done_message=done_message,
            )

        if dry_run:
            progress.live_display(Panel.fit(sbatch_content, title="sbatch file"))
            exit(0)

        with progress.enter_task(f"Uploading sbatch file to {c.original_host}"):
            sbatch_fd = StringIO(sbatch_content)
            c.put(sbatch_fd, remote=sbatch_path)

        with progress.enter_task(f"Making sbatch file executable"):
            c.run(f"chmod +x {sbatch_path}", hide=True)

        with progress.enter_task(f"Submitting job to SLURM queue") as task:
            result = c.run(f"sbatch {sbatch_path}", hide=True, warn=True)
            if result.ok and result.stdout:
                job_id = int(result.stdout.split()[-1])
                progress.update(
                    task, description=f"Job submitted to batch queue. Job ID: {job_id}"
                )
            else:
                raise RuntimeError("Job submission failed.\n" + result.stdout)

        with progress.enter_task(f"Waiting for job to start") as task:
            for _ in range(timeout):
                result = c.run(f"squeue -j {job_id}", hide=True, warn=True)
                df = parse_squeue_output(result.stdout)
                if len(df) and df["ST"].iloc[0] == "R":
                    progress.update(task, description=f"Job started")
                    break
                elif len(df) and df["ST"].iloc[0] == "C":
                    raise RuntimeError("Job was cancelled.")
                time.sleep(0.5)
            else:
                raise RuntimeError("Timeout reached while waiting for job to start.")

        with progress.enter_task(f"Waiting for job to finish") as task:
            for _ in range(hours * 3600 // 2):
                time.sleep(2)
                if not is_file(c, output, hide=True):
                    continue
                result = c.run(f"tail -n 5 {output}", hide=True, warn=True)
                if result.ok and result.stdout:
                    progress.live_display(Panel.fit(result.stdout, title="Output file"))
                if done_message in result.stdout:
                    progress.update(task, description=f"Output file ready")
                    progress.live_display(None)
                    break

        with progress.enter_task(f"Getting output files") as task:
            result = c.run(f"cat {output}", hide=True, warn=True)
            if result.ok and result.stdout:
                out = tail(result.stdout, 50)
                progress.console.print(Panel(out, title="Output file"))
            result = c.run(f"cat {error}", hide=True, warn=True)
            if result.ok and result.stdout:
                err = tail(result.stdout, 50)
                progress.console.print(Panel(err, title="Error file"))



from fabric.connection import Connection
from fabric.tasks import task

from xefab.utils import console

SHELL_PROFILE_FILES = {
    "sh": ["~/.profile"],
    "bash": ["~/.profile", "~/.bash_profile"],
    "zsh": ["~/.profile", "~/.zprofile"],
}


@task(default=True)
def shell(c: Connection, shell: str = None):
    """Open interactive shell on remote host."""

    if shell is None:
        shell = c.config.run.shell
    if isinstance(c, Connection):
        c.shell()
    else:
        while True:
            try:
                cmd = input(f"{shell}:~$ ")
                if cmd == "exit":
                    break
                c.run(cmd, shell=shell)
            except KeyboardInterrupt:
                break


@task
def is_file(c, path: str, local: bool = False, hide: bool = False):
    """Check if a file exists."""

    cmd = f"test -f {path}"
    if local:
        result = c.local(cmd, hide=True, warn=True)
    else:
        result = c.run(cmd, hide=True, warn=True)
    if not hide:
        msg = "1" if result.ok else "0"
        console.print(msg)
    return result.ok


@task
def is_dir(c, path: str, local: bool = False, hide: bool = False):
    """Check if a directory exists."""

    cmd = f"test -d {path}"
    if local:
        result = c.local(cmd, hide=True, warn=True)
    else:
        result = c.run(cmd, hide=True, warn=True)
    if not hide:
        msg = "1" if result.ok else "0"
        console.print(msg)
    return result.ok


@task
def exists(c, path: str, local: bool = False, hide: bool = False):
    """Check if a file or directory exists."""

    cmd = f"test -e {path}"
    if local:
        result = c.local(cmd, hide=hide, warn=True)
    else:
        result = c.run(cmd, hide=hide, warn=True)
    if not hide:
        msg = "1" if result.ok else "0"
        console.print(msg)
    return result.ok


@task
def which(
    c,
    command: str,
    local: bool = False,
    no_user_profile: bool = False,
    no_system_profile: bool = False,
    shell: str = None,
    hide: bool = False,
):
    """Find the path to a command on the remote host."""

    if shell is None:
        shell = "bash"

    cmd = f"which {command}"

    if not no_system_profile and is_file(c, f"/etc/profile", local=local, hide=True):
        cmd = f"source /etc/profile && {cmd}"
    if not no_user_profile:
        for f in SHELL_PROFILE_FILES.get(shell):
            fpath = f.replace("~/", f"/home/{c.user}/")
            if is_file(c, fpath, local=local, hide=True):
                cmd = f"source {fpath} && {cmd}"

    if local and isinstance(c, Connection):
        result = c.local(cmd, hide=True, warn=True, shell=f"/bin/{shell}")
    else:
        result = c.run(cmd, hide=True, warn=True, shell=f"/bin/{shell}")

    if not hide:
        msg = result.stdout.strip() if result.ok else ""
        console.print(msg)
    return result.stdout.strip() if result.ok else None


@task
def get_system(c, hide=False):
    """Deduce the system type of the remote host."""

    result = c.run("python -m platform", hide=True, warn=True)
    assert result.ok, "Failed to deduce system."
    system = result.stdout.split("-")[0]
    if not hide:
        console.print(f"System: {system}")
    return system


@task
def path(
    c,
    profile: bool = False,
    shell: str = None,
    local: bool = False,
    hide: bool = False,
):
    """Print the PATH variable on the remote host."""

    if shell is None:
        shell = "bash"

    cmd = "echo $PATH"

    if local and isinstance(c, Connection):
        result = c.local(cmd, hide=True, warn=True, shell=f"/bin/{shell}")
    else:
        if profile:
            if is_file(c, f"/etc/profile", hide=True):
                cmd = f"source /etc/profile && {cmd}"
            for f in SHELL_PROFILE_FILES.get(shell):
                if is_file(c, f, hide=True):
                    fpath = f.replace("~/", f"/home/{c.user}/")
                    cmd = f"source {fpath} && {cmd}"

        result = c.run(cmd, hide=True, warn=True, shell=f"/bin/{shell}")
    assert (
        result.ok
    ), f"Failed to print path. stdout: {result.stdout} \n stderr: {result.stderr}"
    if not hide:
        console.print(result.stdout.strip())
    return result.stdout.strip()



import six
from fabric.tasks import task

from xefab.utils import console


@task
def download_file(c, path: str, out: str = None, preserve_mode: bool = True):
    """Download a file from a remote server."""

    with console.status(f"Downloading {path} from {c.host}"):
        c.get(path, local=out, preserve_mode=preserve_mode)
    console.print(f"Done.")


@task
def upload_file(c, path: str, remote_path: str = None, preserve_mode: bool = True):
    """Upload a file to a remote server."""

    with console.status(f"Uploading {path} to {c.host}"):
        c.put(path, remote=remote_path, preserve_mode=preserve_mode)
    console.print(f"Done.")


def rsync(
    c,
    source,
    target,
    exclude=(),
    delete=False,
    strict_host_keys=True,
    rsync_opts="",
    ssh_opts="",
):
    """
    Convenient wrapper around your friendly local ``rsync``.
    Copied as is from patchwork since its dependency on fabric 2.x conflicts.

    Specifically, it calls your local ``rsync`` program via a subprocess, and
    fills in its arguments with Fabric's current target host/user/port. It
    provides Python level keyword arguments for some common rsync options, and
    allows you to specify custom options via a string if required (see below.)

    For details on how ``rsync`` works, please see its manpage. ``rsync`` must
    be installed on both the invoking system and the target in order for this
    function to work correctly.

    .. note::
        This function transparently honors the given
        `~fabric.connection.Connection`'s connection parameters such as port
        number and SSH key path.

    .. note::
        For reference, the approximate ``rsync`` command-line call that is
        constructed by this function is the following::

            rsync [--delete] [--exclude exclude[0][, --exclude[1][, ...]]] \\
                -pthrvz [rsync_opts] <source> <host_string>:<target>

    :param c:
        `~fabric.connection.Connection` object upon which to operate.
    :param str source:
        The local path to copy from. Actually a string passed verbatim to
        ``rsync``, and thus may be a single directory (``"my_directory"``) or
        multiple directories (``"dir1 dir2"``). See the ``rsync`` documentation
        for details.
    :param str target:
        The path to sync with on the remote end. Due to how ``rsync`` is
        implemented, the exact behavior depends on the value of ``source``:

        - If ``source`` ends with a trailing slash, the files will be dropped
          inside of ``target``. E.g. ``rsync(c, "foldername/",
          "/home/username/project")`` will drop the contents of ``foldername``
          inside of ``/home/username/project``.
        - If ``source`` does **not** end with a trailing slash, ``target`` is
          effectively the "parent" directory, and a new directory named after
          ``source`` will be created inside of it. So ``rsync(c, "foldername",
          "/home/username")`` would create a new directory
          ``/home/username/foldername`` (if needed) and place the files there.

    :param exclude:
        Optional, may be a single string or an iterable of strings, and is
        used to pass one or more ``--exclude`` options to ``rsync``.
    :param bool delete:
        A boolean controlling whether ``rsync``'s ``--delete`` option is used.
        If True, instructs ``rsync`` to remove remote files that no longer
        exist locally. Defaults to False.
    :param bool strict_host_keys:
        Boolean determining whether to enable/disable the SSH-level option
        ``StrictHostKeyChecking`` (useful for frequently-changing hosts such as
        virtual machines or cloud instances.) Defaults to True.
    :param str rsync_opts:
        An optional, arbitrary string which you may use to pass custom
        arguments or options to ``rsync``.
    :param str ssh_opts:
        Like ``rsync_opts`` but specifically for the SSH options string
        (rsync's ``--rsh`` flag.)
    """
    # Turn single-string exclude into a one-item list for consistency
    if isinstance(exclude, six.string_types):
        exclude = [exclude]
    # Create --exclude options from exclude list
    exclude_opts = ' --exclude "{}"' * len(exclude)
    # Double-backslash-escape
    exclusions = tuple([str(s).replace('"', '\\\\"') for s in exclude])
    # Honor SSH key(s)
    key_string = ""
    # TODO: seems plausible we need to look in multiple places if there's too
    # much deferred evaluation going on in how we eg source SSH config files
    # and so forth, re: connect_kwargs
    # TODO: we could get VERY fancy here by eg generating a tempfile from any
    # in-memory-only keys...but that's also arguably a security risk, so...
    keys = c.connect_kwargs.get("key_filename", [])
    # TODO: would definitely be nice for Connection/FabricConfig to expose an
    # always-a-list, always-up-to-date-from-all-sources attribute to save us
    # from having to do this sort of thing. (may want to wait for Paramiko auth
    # overhaul tho!)
    if isinstance(keys, six.string_types):
        keys = [keys]
    if keys:
        key_string = "-i " + " -i ".join(keys)
    # Get base cxn params
    user, host, port = c.user, c.host, c.port
    port_string = "-p {}".format(port)
    # Remote shell (SSH) options
    rsh_string = ""
    # Strict host key checking
    disable_keys = "-o StrictHostKeyChecking=no"
    if not strict_host_keys and disable_keys not in ssh_opts:
        ssh_opts += " {}".format(disable_keys)
    rsh_parts = [key_string, port_string, ssh_opts]
    if any(rsh_parts):
        rsh_string = "--rsh='ssh {}'".format(" ".join(rsh_parts))
    # Set up options part of string
    options_map = {
        "delete": "--delete" if delete else "",
        "exclude": exclude_opts.format(*exclusions),
        "rsh": rsh_string,
        "extra": rsync_opts,
    }
    options = "{delete}{exclude} -pthrvz {extra} {rsh}".format(**options_map)
    # Create and run final command string
    # TODO: richer host object exposing stuff like .address_is_ipv6 or whatever
    if host.count(":") > 1:
        # Square brackets are mandatory for IPv6 rsync address,
        # even if port number is not specified
        cmd = "rsync {} {} [{}@{}]:{}"
    else:
        cmd = "rsync {} {} {}@{}:{}"
    cmd = cmd.format(options, source, user, host, target)
    return c.local(cmd)



import sys

from decopatch import DECORATED, decorator
from fabric.connection import Connection
from fabric.tasks import task as fabric_task
from invoke.context import DataProxy
from pydantic import BaseModel

from xefab.collection import XefabCollection
from xefab.utils import console

from ..pydantic_support import task_from_model
from . import admin, github, install, secrets, shell

namespace = XefabCollection("root")


@decorator
def task(*args, f=DECORATED, **kwargs):
    """
    If used to decorate a pydantic Model, then create a task from the model.
    If not, then just apply the fabric task decorator.
    """
    if (isinstance(f, type) and issubclass(f, BaseModel)) or isinstance(f, BaseModel):
        task = task_from_model(f, *args, **kwargs)
        namespace.add_task(task)
        return f
    return fabric_task(f, *args, **kwargs)


def printable(d):
    if isinstance(d, (dict, DataProxy)):
        return {k: printable(v) for k, v in d.items()}
    elif isinstance(d, type):
        return d.__qualname__
    return d


@task
def show_context(c, config_name: str = None):
    """Show the context being used for tasks."""

    if config_name is None:
        console.print_json(data=printable(c), indent=4)
        return
    result = getattr(c, config_name, None)
    if result is None:
        result = c.get(config_name, None)
    if result is None:
        console.print(f"Config {config_name} not found.")
    result = printable(result)
    if isinstance(result, dict):
        console.print_json(data=result, indent=4)
    else:
        console.print(f"{config_name}: {result}")


install = XefabCollection.from_module(install, name="install")
namespace.add_collection(install)

secret = XefabCollection.from_module(secrets, name="secrets")
namespace.add_collection(secret)

admin = XefabCollection.from_module(admin, name="admin")
namespace.add_collection(admin)

github = XefabCollection.from_module(github, name="github")
namespace.add_collection(github)

sh = XefabCollection.from_module(shell, name="sh")
namespace.add_collection(sh)


namespace.add_task(show_context)



"""Console script for xesites."""
import inspect
import sys
from collections import defaultdict

# IMPORTANT: Must be imported before fabric for monkey patching to work
# isort: off
from xefab.ssh_client import SSHClient  # isort: skip

# isort: on

from fabric.executor import Executor
from fabric.main import Fab
from invoke.exceptions import Exit
from invoke.parser import Argument
from invoke.util import debug, helpline
from rich.console import Group, NewLine, group
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from xefab import __version__, tasks
from xefab.collection import XefabCollection
from xefab.config import Config
from xefab.utils import ProgressContext, console, nested_dict


@group()
def get_tuples_group(header, docstring, tuples=None):
    """Create a group of tuples."""
    yield Text(header, style="bold")
    yield NewLine()
    if docstring:
        yield Text(docstring, style="italic")
    yield NewLine()
    yield Text("Options: ", style="bold")
    grid = Table.grid(expand=True)
    grid.add_column()
    grid.add_column(justify="left")
    for line in tuples:
        grid.add_row(*line)
    yield grid


def help_tuples_to_grid(tuples):
    """Convert a list of tuples to a rich table."""
    grid = Table.grid(expand=True)
    grid.add_column(justify="left", style="bold")
    grid.add_column(
        justify="left",
        style="dim",
    )
    for line in tuples:
        grid.add_row(*line)
    return grid


class XeFab(Fab):
    """XeFab CLI"""

    ROOT_COLLECTION_NAME = "main"
    USER_COLLECTION_NAME = "my-tasks"

    def core_args(self):
        """Add xefab config to core args."""
        DEFAULTS = {
            "list-depth": 2,
            "list-format": "flat",
        }
        args = super().core_args()

        for arg in args:
            for name in arg.names:
                if name in DEFAULTS:
                    arg.default = DEFAULTS[name]

        my_args = [
            # Argument(
            #     names=("verbose", "v"),
            #     kind=int,
            #     incrementable=True,
            #     default=0,
            #     help="Increase verbosity of outputs.",
            # ),
        ]

        return args + my_args

    def task_args(self):
        return super().task_args()

    def parse_collection(self):
        user_namespace = None
        # Load any locally defined tasks
        if self.namespace is not None:
            user_namespace = self.namespace
        else:
            try:
                self.load_collection()
                user_namespace = self.collection
            except Exit:
                pass

        # Load the default tasks
        self.namespace = XefabCollection.from_module(
            tasks, name=self.ROOT_COLLECTION_NAME
        )
        self.namespace.load_objects_from_entry_points()
        if user_namespace is not None:
            for name, task in user_namespace.tasks.items():
                self.namespace.add_task(task, name=name)
            for name, collection in user_namespace.collections.items():
                self.namespace.add_collection(collection, name=name)

        hostname = None
        if len(self.argv) > 1:
            argv = [self.argv.pop(0)]
            original_argv = list(self.argv)
            while self.argv:
                arg = self.argv.pop(0)
                if arg in ["-h", "--help", "--list", "-v", "--verbose"]:
                    argv.insert(0, arg)
                    continue

                if "." in arg:
                    arg, _, rest = arg.partition(".")
                    self.argv.insert(0, rest)

                if arg in self.namespace.collections:
                    self.namespace = self.namespace.collections[arg]
                    hostnames = self.namespace._configuration.get("hostnames", None)
                    debug(f"xefab: {arg} hostnames: {hostnames}")
                    self.config.configure_ssh_for_host(arg, hostnames)
                    if hostnames is not None and hostname is None:
                        hostname = arg
                else:
                    argv.append(arg)
                    argv.extend(self.argv)
                    break

            if argv != original_argv:
                self.argv = argv
                self.parse_core(argv)

            # if we loaded a host collection, it sets the default host argument
            if not self.args.hosts.value and hostname is not None:
                self.args.hosts.value = hostname
        super().parse_collection()

    def task_panel(self, task, name, parents=None):
        """Create a help panel for a specific task."""
        if parents is None:
            parents = ()
        if isinstance(parents, str):
            parents = (parents,)
        if isinstance(parents, tuple):
            parents = [parents]
        docstring = inspect.getdoc(task)
        if docstring is None:
            docstring = ""
        tuples = []
        if name in self.parser.contexts:
            ctx = self.parser.contexts[name]
            tuples = ctx.help_tuples()
        elif parents:
            for parent in parents:
                if f"{'.'.join(parent)}.{name}" in self.parser.contexts:
                    ctx = self.parser.contexts[f"{'.'.join(parent)}.{name}"]
                    tuples = ctx.help_tuples()
                    break

        if len(parents) > 1:
            header = "{} [--core-opts] <Collection> {} {}[other tasks here ...]"
            host_str = ",".join([" ".join(parent) for parent in parents])
        else:
            header = "{} [--core-opts] {} {}[other tasks here ...]"
            host_str = ""
        options_str = "[--options]" if tuples else ""

        if len(parents) == 1:
            name = " ".join(parents[0] + (name,))
        header = header.format(self.binary, name, options_str)
        if host_str:
            header += f"\nCollection options: {host_str}"
        content = get_tuples_group(header, docstring, tuples=tuples)

        return Panel(
            content,
            title=Text(
                name,
                style="bold",
            ),
            title_align="left",
        )

    def task_tree(self, collection, tree=None, parents=()):
        """Create a tree of tasks."""
        if tree is None:
            tree = Tree("")

        if self.list_depth <= len(parents):
            return tree

        for name, task in collection.tasks.items():
            if self.list_format == "tree":
                panel = self.task_panel(
                    task, name, parents=parents + (collection.name,)
                )
                tree.add(panel)
            else:
                tree.add(name)
        for name, subcollection in collection.collections.items():
            subtree = tree.add(name)
            self.task_tree(subcollection, tree=subtree, parents=parents + (name,))
        return tree

    def unique_tasks(self, collection, parents=()):
        """Create a set of unique tasks."""
        tasks = set()
        for name, task in collection.tasks.items():
            tasks.add(task)

        for name, subcollection in collection.collections.items():
            tasks.update(self.unique_tasks(subcollection, parents=parents + (name,)))
        return tasks

    def find_task_paths(self, task, collection, parents=()):
        """Find all paths to a task."""
        paths = []
        for name, subtask in collection.tasks.items():
            if subtask == task:
                paths.append(parents)

        for name, subcollection in collection.collections.items():
            paths.extend(
                self.find_task_paths(task, subcollection, parents=parents + (name,))
            )
        return paths

    def collection_panel(self, collection, parents=()):
        """Create a help panel for a specific collection."""
        panels = []
        for task in self.unique_tasks(collection):
            task_paths = self.find_task_paths(task, collection)
            panel = self.task_panel(task, task.name, parents=task_paths)
            panels.append(panel)
        return Group(*panels)

    def list_flat(self):
        pairs = self._make_pairs(self.scoped_collection)
        self.display_with_columns(pairs=pairs)

    def _make_help_pairs(self, coll, nested=False):
        pairs = []
        for name, task in sorted(coll.tasks.items()):
            is_default = name == coll.default
            # Start with just the name and just the aliases, no prefixes or
            # dots.
            displayname = name
            aliases = list(map(coll.transform, sorted(task.aliases)))

            # Nested? add asterisks to default-tasks.
            if nested and is_default:
                displayname += "*"

            # Generate full name and help columns and add to pairs.
            alias_str = " ({})".format(", ".join(aliases)) if aliases else ""
            full = (displayname + alias_str).strip()
            help_str = (helpline(task) or "").strip()
            pairs.append((full, help_str))

        return pairs

    def _make_help_tree(self, coll, ancestors=None, tree=None):
        if tree is None:
            tree = Tree("")

        if ancestors is None:
            ancestors = ()

        if not ancestors and not coll.collections:
            grid = help_tuples_to_grid(self._make_help_pairs(coll))
            return grid

        grid = help_tuples_to_grid(self._make_help_pairs(coll))
        tree.add(grid)

        for name, collection in sorted(coll.collections.items()):
            subtree = tree.add(name)
            self._make_help_tree(
                collection, ancestors=ancestors + (coll.name,), tree=subtree
            )

        return tree

    def list_usage(self):
        """List all tasks in the current namespace and their details."""
        console.print(Text(f"\n\nAvailable Tasks:\n", style="bold"))
        console.print(self._make_help_tree(self.scoped_collection))
        console.print(Text(f"\n\nTask Usage:\n", style="bold"))
        console.print(
            self.collection_panel(
                self.scoped_collection,
            )
        )

    def list_nested(self):
        """List all tasks in the current namespace."""
        console.print(Text(f"\n\nAvailable Tasks:\n", style="bold"))
        console.print(self._make_help_tree(self.scoped_collection))

    def print_task_help(self, name):
        """Print help for a specific task."""
        if name in self.collection.tasks:
            task = self.collection.tasks[name]
        console.print(self.task_panel(task, name))

    def print_columns(self, tuples, indent=0):
        """Print a list of tuples in columns."""
        grid = Table.grid(expand=True)
        grid.add_column()
        grid.add_column()
        for line in tuples:
            grid.add_row(*line)
        console.print(grid)

    def print_help(self):
        usage_suffix = (
            "[<host>] [<subcommand>] task1 [--task1-opts] ... taskN [--taskN-opts]"
        )
        if (
            self.namespace is not None
            and self.namespace.name != self.ROOT_COLLECTION_NAME
        ):
            usage_suffix = f"{self.namespace.name} <subcommand> [--subcommand-opts] ..."
        console.print("Usage: {} [--core-opts] {}".format(self.binary, usage_suffix))
        console.print("")
        console.print("Core options:")
        console.print("")
        self.print_columns(self.initial_context.help_tuples())
        console.print("\n")
        if self.namespace is not None:
            self.list_tasks()


def make_program():
    return XeFab(
        name="xefab",
        version=__version__,
        executor_class=Executor,
        config_class=Config,
    )


program = make_program()


if __name__ == "__main__":
    sys.exit(program.run())  # pragma: no cover



import inspect
import json
import re

import toml
import yaml
from fabric.tasks import Task
from invoke.context import Context
from makefun import create_function
from pydantic import BaseModel

from .utils import camel_to_snake, console


def read_file(file_path: str):
    fname, _, ext = file_path.rpartition(".")
    if ext == "json":
        with open(file_path) as f:
            data = json.load(f)
    elif ext in ["yaml", "yml"]:
        with open(file_path) as f:
            data = yaml.safe_load(f)
    elif ext == "toml":
        with open(file_path) as f:
            data = toml.load(f)
    else:
        raise ValueError(f"Unknown file extension {ext}")
    return data


def get_pydantic_signature(model, defaults={}):
    context_param = inspect.Parameter(
        "c",
        inspect.Parameter.POSITIONAL_ONLY,
        annotation=Context,
    )

    file_param = inspect.Parameter(
        "parse_file",
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        annotation=str,
        default=None,
    )

    params = [
        context_param,
        file_param,
    ]
    for name, field in model.__fields__.items():
        alias = field.alias
        for type_ in model.mro():
            if name in getattr(type_, "__annotations__", {}):
                annotation = type_.__annotations__[name]
                default = defaults.get(name, field.default)
                param = inspect.Parameter(
                    name,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=default,
                    annotation=annotation,
                )
                params.append(param)

                if name == alias:
                    break

                alias_param = param = inspect.Parameter(
                    alias,
                    inspect.Parameter.KEYWORD_ONLY,
                    default=default,
                    annotation=annotation,
                )
                params.append(alias_param)
                break

    return inspect.Signature(params)


def get_pydantic_field_help(field, default=None):
    """
    Takes a model and a field name, and returns a string containing the field's title, description,
    and default value

    :param model: The model class
    :param field_name: The name of the field to get the help for
    :return: The help string for the field.
    """
    title = field.field_info.title
    description = field.field_info.description
    default = default or field.default
    help_str = ""
    if title:
        help_str = f"{title}: "
    if description:
        help_str += description
    if default:
        help_str += f" (default: {default})"
    return help_str


def task_from_model(model, *args, **kwargs):
    """
    Create a task from a pydantic model.
    The model must have a run method.
    The run method must accept a single parameter, the Context object.
    """
    model_class = model
    defaults = {}
    if not isinstance(model, type):
        model_class = model.__class__
        defaults = model.dict()
    class_name = model_class.__name__

    if not issubclass(model_class, BaseModel):
        raise TypeError(
            f"Cant create a task from type {model_class} must be a subclass of pydantic.BaseModel"
        )

    if not hasattr(model, "__call__"):
        raise TypeError(
            f"{model} must implement a __call__ method to be used as a task."
        )

    meth = getattr(model_class, "__call__")
    run_signature = inspect.signature(meth)

    if len(run_signature.parameters) != 2:
        raise TypeError(
            f"{class_name}.run must accept exactly two arguments: self and the Context object."
        )

    name = camel_to_snake(class_name)

    signature = get_pydantic_signature(model_class, defaults=defaults)

    def task_implementation(c, parse_file=None, **kwargs):
        if parse_file is not None:
            data = read_file(parse_file)
            cli_kwargs = {
                k: v for k, v in kwargs.items() if v != signature.parameters[k].default
            }
            merged = dict(data, **cli_kwargs)
            kwargs = dict(kwargs, **merged)
        model = model_class(**kwargs)
        return model(c)

    func = create_function(
        signature, task_implementation, func_name=name, doc=inspect.getdoc(meth)
    )

    task_class = kwargs.pop("klass", Task)
    default_help = {
        k: get_pydantic_field_help(field, defaults.get(k, None))
        for k, field in model.__fields__.items()
    }
    kwargs.setdefault("help", default_help)

    if not args:
        return task_class(func, **kwargs)

    # @task(pre, tasks, here)
    if "pre" in kwargs:
        raise TypeError("May not give *args and 'pre' kwarg simultaneously!")
    kwargs["pre"] = args
    return task_class(func, **kwargs)



import pandas as pd
from fabric.tasks import task

from xefab.utils import console


def add_recipient(
    c, key: dict, gnuhome: str = "~/.gnupg", gopass_store: str = "xenonnt"
):
    """Add a recipient to the gopass store."""
    import gnupg

    gpg = gnupg.GPG(homedir=gnuhome)
    gpg.import_keys(key["raw_key"])
    console.print(f"Added {key['key_id']} to {gnuhome}")
    result = c.run(
        f"gopass --yes recipients add --store {gopass_store} {key['key_id']}",
        hide=True,
        warn=True,
    )
    if result.ok:
        console.print(f"Added {key['key_id']} to gopass")
    else:
        console.print(f"Failed to add {key['key_id']} to gopass: {result.stderr}")


@task
def user_db(c, limit: int = None, hide: bool = False):
    """Get all users from the user database."""
    users = c.config.xent_collection(collection="users").find({}, projection={"_id": 0})
    if limit is not None:
        users = users.limit(int(limit))
    df = pd.json_normalize(list(users))
    if not hide:
        console.print_json(df.to_json(orient="records"), indent=4)
    return df



import json
import time
from io import StringIO

from fabric.tasks import task
from invoke.watchers import Responder
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.text import Text

from xefab.utils import ProgressContext, console

from .github import clone
from .shell import exists
from .utils import print_splash


@task(pre=[print_splash])
def mc_chain(
    c,
    simulation_name: str = "Pmt_neutron",
    start_job: int = 0,
    num_events: int = 1000,
    jobs_per_batch: int = 5,
    events_per_job: int = 100,
    experiment: str = "XENONnT",
    mc_version: str = "head",
    mc_preinit_macro: str = "preinit_nVeto.mac",
    mc_preinit_belt: str = "preinit_B_none.mac",
    mc_optical_setup: str = "setup_optical.mac",
    mc_source_macro: str = "run_Cryostat_neutron.mac",
    run_wfsim: bool = False,
    epix_detectorconfig: str = "sr0_epix_detectorconfig.ini",
    epix_microseparation: float = 0.05,
    epix_tagclusterby: str = "energy",
    epix_nr_only: bool = False,
    event_rate: float = 1,
    chunk_size: int = 100,
    save_raw_records: bool = False,
    run_fast_sim: bool = False,
    sim_nv: bool = False,
    config_file: str = None,
):
    """Run a full chain MC simulation"""

    if config_file is None:
        config = {
            "simulation_name": simulation_name,
            "start_job": start_job,
            "num_events": num_events,
            "jobs_per_batch": jobs_per_batch,
            "events_per_job": events_per_job,
            "experiment": experiment,
            "mc_version": mc_version,
            "mc_preinit_macro": mc_preinit_macro,
            "mc_preinit_belt": mc_preinit_belt,
            "mc_optical_setup": mc_optical_setup,
            "mc_source_macro": mc_source_macro,
            "run_wfsim": run_wfsim,
            "epix_detectorconfig": epix_detectorconfig,
            "epix_microseparation": epix_microseparation,
            "epix_tagclusterby": epix_tagclusterby,
            "epix_nr_only": epix_nr_only,
            "event_rate": event_rate,
            "chunk_size": chunk_size,
            "save_raw_records": save_raw_records,
            "run_fast_sim": run_fast_sim,
            "sim_nv": sim_nv,
        }

        config_file = StringIO(json.dumps(config, indent=4))

    PRODUCTION_DIR = f"/scratch/{c.user}/simulations"
    repo_dir = f"{PRODUCTION_DIR}/mc_chain"
    config_file_remote_path = f"{repo_dir}/{simulation_name}_config.json"
    with ProgressContext() as progress:
        # Check if the production directory exists
        if not exists(c, repo_dir, hide=True):
            # Create the production directory
            with progress.enter_task("Creating production directory"):
                c.run(f"mkdir -p {PRODUCTION_DIR}", hide=True)

            # Clone the MC repo
            with progress.enter_task("Cloning MC repo"):
                clone(c, repo="mc_chain", dest=repo_dir, hide=True)

        # Copy the config file
        with progress.enter_task("Uploading config file"):
            r = c.put(config_file, remote=config_file_remote_path)

        # FIXME: Do we need to renew the user proxy? or just use the shared xenon one?
        # with progress.enter_task("Renewing user certificate proxy"):
        #     password = console.input("Enter GRID pass phrase for this identity:")
        #     responder = Responder(
        #             pattern=r"Enter GRID pass phrase for this identity:",
        #             response=password,
        #             )
        #     c.run("voms-proxy-init -voms xenon.biggrid.nl -bits 2048 -hours 168 --debug -out ~/user_cert",
        #             hide=True, watchers=[responder])

        # Switch to the repo directory
        with c.cd(repo_dir):
            # Setup the environment
            with c.prefix("source setup_env.sh"):
                # Run the chain
                with progress.enter_task("Bulding and submiting workflow"):
                    result = c.run(
                        f"python mc_chain.py --config {config_file_remote_path} --skip-rucio",
                        out_stream=progress.console.file,
                        hide=False,
                    )
                    if result.failed:
                        console.print("Error while building and submitting workflow")
                        console.print(result.stdout)
                        return
        status_command = None
        remove_command = None
        for line in result.stdout.splitlines():
            if "pegasus-status" in line:
                status_command = line.strip()
            if "pegasus-remove" in line:
                remove_command = line.strip()
        try:
            msg = Text("Press Ctrl+C to exit.", style="bold red")
            with progress.enter_task(
                "Monitoring status",
            ):
                while True:
                    result = c.run(status_command, hide=True)
                    status = Panel.fit(result.stdout, title="Workflow Status")
                    display = Group(status, msg)
                    progress.live_display(display)
                    for line in result.stdout.splitlines():
                        if "no matching jobs found in Condor Q" in line:
                            exit(0)
                    time.sleep(2)
        except KeyboardInterrupt:
            pass

        except Exception as e:
            console.print(e)
    if remove_command is not None:
        resp = console.input("Cancel workflow? \[y/N]: ")
        if resp.lower() in ["y", "yes"]:
            with c.cd(repo_dir):
                with c.prefix("source setup_env.sh"):
                    c.run(f"{remove_command}")
    console.print("Goodbye!")



import json
import os
import tempfile

from fabric.connection import Connection
from fabric.tasks import task

from xefab.collection import XefabCollection
from xefab.utils import console, filesystem, try_local

from .admin import add_recipient
from .install import ensure_dependency, github_cli
from .shell import which
from .transfer import rsync

namespace = XefabCollection("github")


def github_api_call(
    c,
    path: str,
    method: str = "GET",
    raw: bool = False,
    hide: bool = False,
    pages: int = 10,
    per_page=50,
):
    """Call the github api and return the result as a list of json documents."""
    promises = []
    for page in range(1, pages + 1):
        cmd = f"gh api {path}?page={page}&per_page={per_page}"
        if method != "GET":
            cmd += f" -X {method}"

        promise = c.run(cmd, hide=True, warn=True, asynchronous=True)
        promises.append(promise)

    for promise in promises:
        result = promise.join()
        if result.failed:
            raise RuntimeError(f"Failed to get github keys: {result.stderr}")

        data = result.stdout
        if data == "[]":
            continue
        if raw:
            yield data
        else:
            yield from json.loads(data)


@task
def is_public(c, repo: str, owner: str = "XENONnT"):
    """Check if a repo is public."""

    for doc in github_api_call(c, f"/users/{owner}/repos", hide=True):
        if doc["name"] == repo:
            return not doc["private"]
    return False


@task
def is_private(c, repo: str, owner: str = "XENONnT"):
    """Check if a repo is private."""

    for doc in github_api_call(c, f"/users/{owner}/repos", hide=True):
        if doc["name"] == repo:
            return doc["private"]
    return True


@task
@try_local
def token(c):
    """Get the github token for the current user."""
    result = c.run(f"gh auth token", hide=True, warn=True)
    if result.failed:
        raise RuntimeError(f"Failed to get github token: {result.stderr}")

    token = result.stdout.strip()
    console.print(token)
    return token


@task(pre=[ensure_dependency("gh", installer=github_cli)])
@try_local
def xenonnt_members(c, hide: bool = False):
    """List all members of the XENONnT organization."""
    users = github_api_call(c, "orgs/XENONnT/members", hide=True)

    usernames = [user["login"] for user in users]
    if not hide:
        console.print(usernames)
        console.print(f"Found {len(usernames)} members.")
    return usernames


@task(pre=[ensure_dependency("gh", installer=github_cli)])
@try_local
def xenon1t_members(c, hide: bool = False):
    """List all members of the XENON1T organization."""
    users = github_api_call(c, "orgs/XENON1T/members", hide=True)

    usernames = [user["login"] for user in users]
    if not hide:
        console.print(usernames)
        console.print(f"Found {len(usernames)} members.")
    return usernames


@task(pre=[ensure_dependency("gh", installer=github_cli)])
@try_local
def xenonnt_keys(c, hide: bool = False, add: bool = False):
    """Get all public keys of all members of the XENONnT organization."""
    users = github_api_call(c, "orgs/XENONnT/members", hide=True)

    keys = []
    for user in users:
        username = user["login"]
        keys = github_api_call(c, f"users/{username}/gpg_keys", hide=True, pages=1)
        keys = [key for key in keys if key["can_encrypt_storage"] and key["raw_key"]]
        if not keys:
            continue
        keys.extend(keys)
        if not hide:
            console.print(f"Found {len(keys)} keys for {username}:")
            for key in keys:
                console.print(key["key_id"])
        if add:
            for key in keys:
                add_recipient(c, key)
    if add:
        c.run("gopass sync", warn=True)
    return keys


@task
def clone(c, repo: str, org="XENONnT", dest: str = None, hide: bool = False):
    """Clone a github repository.
    If the local host has the `gh` cli installed, it will be used to clone
     the repository and the folder will then be rsynced to the remote host.
    Alternatively, if the remote host has the `gh` cli installed, it will be
    used to clone the repository directly into the target dir.
    Otherwise, the git cli will be used on the host."""
    if dest is None:
        dest = repo

    repo_fullname = f"{org}/{repo}"

    if isinstance(c, Connection) and which(c, "gh", local=True, hide=True):
        with tempfile.TemporaryDirectory() as tmpdir:
            local_path = os.path.join(tmpdir, repo)
            c.local(f"gh repo clone {repo_fullname} {local_path}", hide=hide)
            if not hide:
                console.print(
                    f"Cloned {repo_fullname} to {local_path}. Copying files to remote Host."
                )

            if not dest.startswith("/") and not dest.startswith("~"):
                dest = f"/home/{c.user}/{dest}"
            if dest.endswith(f"/{repo}"):
                dest = dest[: -len(repo)]
            if dest.endswith(f"/{repo}/"):
                dest = dest[: -len(repo) - 1]
            rsync(c, source=local_path, target=dest)

    elif which(c, "gh", hide=True):
        c.run(f"gh repo clone {repo_fullname} {dest}", hide=hide)
    else:
        c.run(f"git clone git@github.com:{repo_fullname}.git {dest}", hide=hide)
    if not hide:
        console.print(f"Cloned {repo_fullname} to {c.user}@{c.host}:{dest}.")


@task
def latest_release_tag(c, repo: str, org: str = "XENONnT", hide: bool = False):
    """Get the latest release tag from a github repository."""
    with c.prefix("export GH_PAGER=cat"):
        if isinstance(c, Connection) and which(c, "gh", local=True, hide=True):
            result = c.local(
                f"gh api repos/{org}/{repo}/releases/latest", hide=True, warn=False
            )
        elif which(c, "gh", hide=True):
            result = c.run(
                f"gh api repos/{org}/{repo}/releases/latest", hide=True, warn=False
            )
        else:
            raise ValueError("No github cli found.")
    release = json.loads(result.stdout)
    tag = release["tag_name"]
    if not hide:
        console.print(tag)
    return tag


namespace.add_task(clone)
namespace.add_task(is_private)
namespace.add_task(is_public)
namespace.add_task(token)
namespace.add_task(xenon1t_members)
namespace.add_task(xenonnt_keys)
namespace.add_task(xenonnt_members)
namespace.add_task(latest_release_tag)



from fabric.tasks import task
from rich.panel import Panel

from xefab.utils import console

UCHICAGO_SPLASH_SCREEN = r"""
 __   __ ______  _   _   ____   _   _      _______ 
 \ \ / /|  ____|| \ | | / __ \ | \ | |    |__   __|
  \ V / | |__   |  \| || |  | ||  \| | _ __  | |   
   > <  |  __|  | . ` || |  | || . ` || '_ \ | |   
  / . \ | |____ | |\  || |__| || |\  || | | || |   
 /_/ \_\|______||_| \_| \____/ |_| \_||_| |_||_|   

                    The UChicago Analysis Center

"""

OSG_SPLASH_SCREEN = r"""
     OOOOOOOOO        SSSSSSSSSSSSSSS         GGGGGGGGGGGGG
   OO:::::::::OO    SS:::::::::::::::S     GGG::::::::::::G
 OO:::::::::::::OO S:::::SSSSSS::::::S   GG:::::::::::::::G
O:::::::OOO:::::::OS:::::S     SSSSSSS  G:::::GGGGGGGG::::G
O::::::O   O::::::OS:::::S             G:::::G       GGGGGG
O:::::O     O:::::OS:::::S            G:::::G              
O:::::O     O:::::O S::::SSSS         G:::::G              
O:::::O     O:::::O  SS::::::SSSSS    G:::::G    GGGGGGGGGG
O:::::O     O:::::O    SSS::::::::SS  G:::::G    G::::::::G
O:::::O     O:::::O       SSSSSS::::S G:::::G    GGGGG::::G
O:::::O     O:::::O            S:::::SG:::::G        G::::G
O::::::O   O::::::O            S:::::S G:::::G       G::::G
O:::::::OOO:::::::OSSSSSSS     S:::::S  G:::::GGGGGGGG::::G
 OO:::::::::::::OO S::::::SSSSSS:::::S   GG:::::::::::::::G
   OO:::::::::OO   S:::::::::::::::SS      GGG::::::GGG:::G
     OOOOOOOOO      SSSSSSSSSSSSSSS           GGGGGG   GGGG

-----------------------------------------------------------
                              Open Science Grid: CIထ nnect
                                                           
"""

SPLASH_SCREENS = {
    "uchicago": UCHICAGO_SPLASH_SCREEN,
    "osg": OSG_SPLASH_SCREEN,
    "dali": UCHICAGO_SPLASH_SCREEN,
    "midway": UCHICAGO_SPLASH_SCREEN,
    "dali-login2.rcc.uchicago.edu": UCHICAGO_SPLASH_SCREEN,
    "dali-login1.rcc.uchicago.edu": UCHICAGO_SPLASH_SCREEN,
    "login.xenon.ci-connect.net": OSG_SPLASH_SCREEN,
    "midway2.rcc.uchicago.edu": UCHICAGO_SPLASH_SCREEN,
    "midway2-login1.rcc.uchicago.edu": UCHICAGO_SPLASH_SCREEN,
    "midway2-login2.rcc.uchicago.edu": UCHICAGO_SPLASH_SCREEN,
}


@task
def print_splash(c):
    """Print the splash screen."""

    if hasattr(c, "host"):
        if c.host in SPLASH_SCREENS:
            console.print(Panel.fit(SPLASH_SCREENS[c.host], title="Welcome"))
            return
    hostnames = getattr(c, "hostnames", [])
    if isinstance(hostnames, str):
        hostnames = hostnames.split(",")
    for host in hostnames:
        if host in SPLASH_SCREENS:
            console.print(Panel.fit(SPLASH_SCREENS[host], title="Welcome"))
            break



from xefab.collection import XefabCollection
from xefab.tasks.condorq import condorq
from xefab.tasks.mc_chain import mc_chain

namespace = XefabCollection("osg")

namespace.configure({"hostnames": "login.xenon.ci-connect.net"})

namespace.add_task(condorq)
namespace.add_task(mc_chain)



from xefab.collection import XefabCollection
from xefab.tasks.batchq import sbatch
from xefab.tasks.jupyter import start_jupyter
from xefab.tasks.main import show_context
from xefab.tasks.squeue import squeue
from xefab.tasks.transfer import download_file, upload_file

namespace = XefabCollection("dali")

namespace.configure(
    {"hostnames": ["dali-login2.rcc.uchicago.edu", "dali-login1.rcc.uchicago.edu"]}
)

namespace.add_task(squeue)
namespace.add_task(start_jupyter)
namespace.add_task(download_file)
namespace.add_task(upload_file)
namespace.add_task(sbatch)
namespace.add_task(show_context)



import datetime

import pandas as pd
from fabric.tasks import task

from xefab.utils import console, df_to_table


@task(aliases=["job-queue"])
def condorq(c, all: bool = False, hide: bool = False):
    """Get the condor job queue."""
    cmd = "condor_q"
    if all:
        cmd += " -allusers"
    result = c.run(cmd, hide=True, warn=True)
    if result.failed:
        console.print(
            f"Remote execution of {cmd} on {c.host} failed. stderr: {result.stderr}"
        )
    df = parse_condorq_output(result.stdout)
    if not hide:
        table = df_to_table(df)
        console.print(table)
    return df


slots = {
    "OWNER": 1,
    "BATCH_NAME": 1,
    "SUBMITTED": 2,
    "DONE": 1,
    "RUN": 1,
    "IDLE": 1,
    "HOLD": 1,
    "TOTAL": 1,
    "JOB_IDS": 3,
}


mergers = {
    "SUBMITTED": lambda x: pd.to_datetime(
        f"{datetime.datetime.utcnow().year}/{x[0]}T{x[1]}"
    ),
    "JOB_IDS": lambda x: "".join(x),
}


def parse_row(line, columns):
    parts = line.split()
    row = {}
    for column in columns:
        value = [parts.pop(0) for _ in range(slots[column])]
        merger = mergers.get(column, lambda x: x[0])
        row[column] = merger(value)

    return row


def parse_condorq_output(condorq_output):
    lines = condorq_output.split("\n")
    rows = []
    columns = []
    for line in lines:
        if line.startswith("OWNER"):
            columns = line.split()
            continue

        if not columns:
            continue

        if not line:
            break

        row = parse_row(line, columns)
        rows.append(row)

    df = pd.DataFrame(rows, columns=columns).dropna(how="all")
    return df



import contextlib
import io
import os
from pathlib import Path

from fabric.connection import Connection
from invoke.context import Context
from pydantic import BaseModel, root_validator

from .utils import console, filesystem


class TemplateFile(BaseModel):
    """
    A template file that can be rendered using a pydantic model.
    The model should contain a __TEMPLATE__ attribute that is a
    string containing the template or a path to a file.
    All fields in the model will be available to the template.
    """

    __NAME__: str = None
    __TEMPLATE__: str = None

    @classmethod
    def get_template(cls):
        path = Path(cls.__TEMPLATE__)
        if path.is_file():
            return path.read_text()
        else:
            return cls.__TEMPLATE__

    @root_validator
    def format_args(cls, values):
        for name, field in cls.__fields__.items():
            value = values.get(name, field.default)
            if isinstance(value, str):
                values[name] = value.format(**values)
        return values

    def __init_subclass__(cls):
        if cls.__TEMPLATE__ is None or not cls.__TEMPLATE__.strip():
            raise TypeError(
                "Template class must have a non-empty __TEMPLATE__ attribute defined"
            )
        if cls.__NAME__ is None or not cls.__NAME__.strip():
            raise TypeError(
                "Template class must have a non-empty __NAME__ attribute defined"
            )

    @property
    def content(self):
        template = self.get_template()
        kwargs = self.dict()
        return template.format(**kwargs)

    @contextlib.contextmanager
    def open(self, mode="r"):
        if mode == "r":
            yield io.StringIO(self.content)
        elif mode == "rb":
            yield io.BytesIO(self.content.encode())
        else:
            raise RuntimeError("File is not writable")

    def deploy(self, c: Context, path, hide=False, warn=False):
        try:
            fs = filesystem(c)
            fs.makedirs(path, exist_ok=True)
            full_path = fs.sep.join([path, self.__NAME__])
            fs.write_text(full_path, self.content)
        except Exception as e:
            if not warn:
                raise e
            if not hide:
                console.print_exception(show_locals=True)

        if not hide:
            console.print(f"Deployed {self.__NAME__} to {path}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tests', 'xefab', 'xefab.hosts', 'xefab.hosts.uchicago', 'xefab.tasks']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'decopatch>=1.4.10,<2.0.0',
 'fabric>=3.0.0,<4.0.0',
 'fsspec>=2023.1.0,<2024.0.0',
 'gnupg>=2.3.1,<3.0.0',
 'makefun>=1.15.0,<2.0.0',
 'pandas',
 'pydantic>=1.10.4,<2.0.0',
 'pymongo',
 'rich>=13.3.1,<14.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['xefab = xefab.main:program.run',
                     'xf = xefab.main:program.run'],
 'xefab.tasks': ['dali = xefab.hosts.uchicago.dali',
                 'midway = xefab.hosts.uchicago.midway',
                 'osg = xefab.hosts.osg']}

setup_kwargs = {
    'name': 'xefab',
    'version': '0.1.5',
    'description': 'Top-level package for xefab.',
    'long_description': "=====\nxefab\n=====\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n\nFabric based task execution for the XENON dark matter experiment\n\n\nInstallation\n------------\n\nTo install xefab, its recomended to use pipx_:\n\n.. code-block:: console\n\n    $ pipx install xefab\n\nAlternatively you can install it using pip:\n\n.. code-block:: console\n\n    $ pip install xefab\n\nUsage\n-----\n\nYou can list the available tasks and options by running xf/xefab without any options.\n\n.. code-block:: console\n\n    $ xf\n    Usage: xf [--core-opts] [<host>] [<subcommand>] task1 [--task1-opts] ... taskN [--taskN-opts]\n\n    Core options:\n\n    --complete                      Print tab-completion candidates for given parse remainder.                                       \n    --hide=STRING                   Set default value of run()'s 'hide' kwarg.                                                       \n    --print-completion-script=STRINGPrint the tab-completion script for your preferred shell (bash|zsh|fish).                        \n    --prompt-for-login-password     Request an upfront SSH-auth password prompt.                                                     \n    --prompt-for-passphrase         Request an upfront SSH key passphrase prompt.                                                    \n    --prompt-for-sudo-password      Prompt user at start of session for the sudo.password config value.                              \n    --write-pyc                     Enable creation of .pyc files.                                                                   \n    -d, --debug                     Enable debug output.                                                                             \n    -D INT, --list-depth=INT        When listing tasks, only show the first INT levels.                                              \n    -e, --echo                      Echo executed commands before running.                                                           \n    -f STRING, --config=STRING      Runtime configuration file to use.                                                               \n    -F STRING, --list-format=STRING Change the display format used when listing tasks. Should be one of: flat (default), nested,     \n                                    json.                                                                                            \n    -h [STRING], --help[=STRING]    Show core or per-task help and exit.                                                             \n    -H STRING, --hosts=STRING       Comma-separated host name(s) to execute tasks against.                                           \n    -i, --identity                  Path to runtime SSH identity (key) file. May be given multiple times.                            \n    -l [STRING], --list[=STRING]    List available tasks, optionally limited to a namespace.                                         \n    -p, --pty                       Use a pty when executing shell commands.                                                         \n    -R, --dry                       Echo commands instead of running.                                                                \n    -S STRING, --ssh-config=STRING  Path to runtime SSH config file.                                                                 \n    -t INT, --connect-timeout=INT   Specifies default connection timeout, in seconds.                                                \n    -T INT, --command-timeout=INT   Specify a global command execution timeout, in seconds.                                          \n    -V, --version                   Show version and exit.                                                                           \n    -w, --warn-only                 Warn, instead of failing, when shell commands fail.                                              \n\n\n    Subcommands:\n\n    show-context                      Show the context being used for tasks.                                                         \n    admin.user-db                                                                                                                    \n    dali.download-file                Download a file from a remote server.                                                          \n    dali.sbatch                       Create and submit a job to SLURM job queue on the remote host.                                 \n    dali.show-context                 Show the context being used for tasks.                                                         \n    dali.squeue (dali.job-queue)      Get the job-queue status.                                                                      \n    dali.start-jupyter                Start a jupyter analysis notebook on the remote host and forward to local port via ssh-tunnel. \n    dali.upload-file                  Upload a file to a remote server.                                                              \n    github.clone                                                                                                                     \n    github.is-private                                                                                                                \n    github.is-public                                                                                                                 \n    github.token                                                                                                                     \n    github.xenon1t-members                                                                                                           \n    github.xenonnt-keys                                                                                                              \n    github.xenonnt-members                                                                                                           \n    install.chezmoi                                                                                                                  \n    install.get-system                                                                                                               \n    install.github-cli (install.gh)                                                                                                  \n    install.gnupg (install.gpg)                                                                                                      \n    install.go                                                                                                                       \n    install.gopass                                                                                                                   \n    install.miniconda (install.conda)                                                                                                \n    install.which                                                                                                                    \n    midway.download-file              Download a file from a remote server.                                                          \n    midway.sbatch                     Create and submit a job to SLURM job queue on the remote host.                                 \n    midway.show-context               Show the context being used for tasks.                                                         \n    midway.squeue (midway.job-queue)  Get the job-queue status.                                                                      \n    midway.start-jupyter              Start a jupyter analysis notebook on the remote host and forward to local port via ssh-tunnel. \n    midway.upload-file                Upload a file to a remote server.                                                              \n    osg.condorq (osg.job-queue)                                                                                                      \n    osg.mc-chain                      Run a full chain MC simulation                                                                 \n    secrets.setup                                                                                                                    \n    secrets.setup-utilix-config                                                                                                      \n    sh.exists                                                                                                                        \n    sh.get-system                                                                                                                    \n    sh.is-dir                                                                                                                        \n    sh.is-file                                                                                                                       \n    sh.path                                                                                                                          \n    sh.shell (sh)                     Open interactive shell on remote host.                                                         \n    sh.which                                                                   \n\nYou can get help for a specific task by running e.g.\n\n.. code-block:: console\n\n    $ xf --help midway.start-jupyter\n    ╭─ start-jupyter ───────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n    │ xf [--core-opts] start-jupyter [--options][other tasks here ...]                                                              │\n    │                                                                                                                               │\n    │ Start a jupyter analysis notebook on the remote host and forward to local port via ssh-tunnel.                                │\n    │                                                                                                                               │\n    │ Options:                                                                                                                      │\n    │ --image-dir=STRING                              Directory to look for singularity images                                      │\n    │ --remote-port=STRING                            Port to use for jupyter server to on the worker node                          │\n    │ --=INT, --local-port=INT                        Local port to attempt to forward to (if free)                                 │\n    │ -a INT, --max-hours=INT                         Maximum number of hours to run for                                            │\n    │ -b, --bypass-reservation                        Dont attempt to use the xenon notebook reservation                            │\n    │ -c INT, --cpu=INT                               Number of CPUs to request                                                     │\n    │ -d, --detached                                  Run the job and exit, dont perform cleanup tasks.                             │\n    │ -e STRING, --env=STRING                         Environment to run on                                                         │\n    │ -f, --force-new                                 Force a new job to be started                                                 │\n    │ -g, --gpu                                       Use a GPU                                                                     │\n    │ -i STRING, --binds=STRING                       Directories to bind to the container                                          │\n    │ -j STRING, --jupyter=STRING                     Type of jupyter server to start (lab or notebook)                             │\n    │ -l, --local-cutax                               Use user installed cutax (from ~/.local)                                      │\n    │ -m INT, --timeout=INT                           Timeout for the job to start                                                  │\n    │ -n STRING, --node=STRING                        Node to run on                                                                │\n    │ -o STRING, --notebook-dir=STRING                Directory to start the notebook in                                            │\n    │ -p STRING, --partition=STRING                   Partition to run on (xenon1t or dali)                                         │\n    │ -r INT, --ram=INT                               Amount of RAM to allocate (in MB)                                             │\n    │ -t STRING, --tag=STRING                         Tag of the container to use                                                   │\n    │ -u, --debug                                     Print debug information                                                       │\n    │ -w, --no-browser                                Dont open the browser automatically when done                                 │\n    ╰───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n\nSome tasks are registered to run on a specific host. When you run them, the --hosts option will be ignored.\n\ne.g. if you run\n\n.. code-block:: console\n\n    $ xf midway start-jupyter\n\nThe task will be run on the midway host, not the host you specified with --hosts.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n.. _pipx: https://github.com/pypa/pipx\n",
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jmosbacher/xefab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import configparser
import os
from pathlib import Path

import appdirs
from fabric.config import Config as FabricConfig
from fabric.config import merge_dicts
from invoke.util import debug
from rich.console import Console

from xefab.entrypoints import get_entry_points
from xefab.utils import console

dirs = appdirs.AppDirs("xefab")

XEFAB_CONFIG = os.getenv(
    "XEFAB_CONFIG", os.path.join(dirs.user_config_dir, "config.env")
)


class Config(FabricConfig):
    """Settings for xefab."""

    MONGO_CLIENTS = {}

    prefix = "xefab"
    xenon_config = None

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("system_prefix", dirs.site_config_dir + "/")
        kwargs.setdefault("user_prefix", dirs.user_config_dir + "/")
        super().__init__(*args, **kwargs)
        self.load_xenon_config()

    def _get_ssh_config(self, hostname):
        """Look up the host in the SSH config, if it exists."""
        config = {"hostname": hostname}

        for host in self.base_ssh_config.get_hostnames():
            data = self.base_ssh_config.lookup(host)
            host_hostname = data.get("hostname", "")
            if hostname in [host, host_hostname]:
                config.update(data)
                return config

    def configure_ssh_for_host(self, host, hostnames=None):
        """Find the SSH config for a host."""
        if isinstance(hostnames, str):
            hostnames = hostnames.split(",")

        config = None

        if hostnames is None:
            return

        if not isinstance(hostnames, list):
            debug(
                "xefab: hostnames must be a list or a string,"
                f" got {type(hostnames)} for host {host}. Ignoring"
            )
            return

        self.load_ssh_config()

        for hostname in hostnames:
            config = self._get_ssh_config(hostname)
            if config is not None:
                debug(f"xefab: found ssh config for {hostname}")
                break
        else:
            config = {"hostname": hostname}

        if "user" not in config:
            config["user"] = console.input(f"Enter username for {hostname}: ")

        ssh_config = {"host": host, "config": config}

        self.base_ssh_config._config.insert(0, ssh_config)

    @staticmethod
    def global_defaults():
        """Add extra parameters to the default dict."""

        defaults = FabricConfig.global_defaults()

        for ep in get_entry_points("xefab.config"):
            try:
                cfg = ep.load()
                if isinstance(cfg, dict):
                    merge_dicts(defaults, cfg)
            except Exception as e:
                debug(f"xefab: Error loading config from {ep.name}: {e}")
                continue

        ours = {
            "tasks": {
                "collection_name": "xetasks",
            },
            "xenon_config_paths": Config.get_xenon_config_paths(),
            "list-depth": 3,
        }

        merge_dicts(defaults, ours)

        return defaults

    @staticmethod
    def get_xenon_config_paths():
        """Get the paths to the xenon config files.
        Files are read in order, so the last one takes precedence.
        """

        paths = [
            os.path.join(dirs.site_config_dir, ".xenon_config"),
            os.path.join(dirs.user_config_dir, ".xenon_config"),
            os.path.join(os.getcwd(), ".xenon_config"),
            os.getenv("XENON_CONFIG", "~/.xenon_config"),
        ]

        return paths

    def load_xenon_config(self):
        """Load the xenon config file."""
        self.xenon_config = configparser.ConfigParser()
        self.merge()  # Make sure we have the latest config
        paths = getattr(self, "xenon_config_paths", [])
        if isinstance(paths, str):
            paths = paths.split(",")
        if not isinstance(paths, list):
            raise ValueError("xenon_config_paths must be a list or a string")
        paths = [os.path.expanduser(path) for path in paths]
        loaded = self.xenon_config.read(paths)
        debug(f"xefab: loaded xenon config from {loaded}")

    def _mongo_client(self, experiment, url=None, user=None, password=None):
        import pymongo

        if experiment not in ["xe1t", "xent"]:
            raise ValueError(
                f"experiment must be 'xe1t' or 'xent'. You passed f{experiment}"
            )

        if not url:
            url = self.xenon_config.get("RunDB", f"{experiment}_url")
        if not user:
            user = self.xenon_config.get("RunDB", f"{experiment}_user")
        if not password:
            password = self.xenon_config.get("RunDB", f"{experiment}_password")

        # build other client kwargs
        max_pool_size = self.xenon_config.get("RunDB", "max_pool_size", fallback=100)
        socket_timeout = self.xenon_config.get(
            "RunDB", "socket_timeout", fallback=60000
        )
        connect_timeout = self.xenon_config.get(
            "RunDB", "connect_timeout", fallback=60000
        )

        uri = f"mongodb://{user}:{password}@{url}"
        if uri not in self.MONGO_CLIENTS:
            self.MONGO_CLIENTS[uri] = pymongo.MongoClient(
                uri,
                readPreference="secondaryPreferred",
                maxPoolSize=max_pool_size,
                socketTimeoutMS=socket_timeout,
                connectTimeoutMS=connect_timeout,
            )
        return self.MONGO_CLIENTS[uri]

    def _mongo_database(self, experiment, database=None, **kwargs):
        if not database:
            database = self.xenon_config.get("RunDB", f"{experiment}_database")
        client = self._mongo_client(experiment, **kwargs)
        return client[database]

    def _mongo_collection(self, experiment, collection, **kwargs):
        client = self._mongo_database(experiment, **kwargs)
        return client[collection]

    def xent_collection(self, collection="runs", **kwargs):
        return self._mongo_collection("xent", collection, **kwargs)

    def xe1t_collection(self, collection="runs_new", **kwargs):
        return self._mongo_collection("xe1t", collection, **kwargs)



import paramiko
from paramiko.ssh_exception import SSHException


class SSHClient(paramiko.SSHClient):
    password = None

    def _2fa_handler(self, title, instructions, prompt_list):
        if not prompt_list:
            return []
        if "Password:" in prompt_list[0][0] and self.password is not None:
            return [self.password]
        if "Duo two-factor login" in prompt_list[0][0]:
            return ["1"]

    def _auth(
        self,
        username,
        password,
        pkey,
        key_filenames,
        allow_agent,
        look_for_keys,
        gss_auth,
        gss_kex,
        gss_deleg_creds,
        gss_host,
        passphrase,
    ):
        """
        Try, in order:
            - The key(s) passed in, if one was passed in.
            - Any key we can find through an SSH agent (if allowed).
            - Any "id_rsa", "id_dsa" or "id_ecdsa" key discoverable in ~/.ssh/
              (if allowed).
            - Plain username/password auth, if a password was given.
        (The password might be needed to unlock a private key [if 'passphrase'
        isn't also given], or for two-factor authentication [for which it is
        required].)
        """

        self.password = password
        if self.password is not None:
            try:
                self._transport.auth_interactive(username, self._2fa_handler)
                return
            except SSHException as e:
                pass

        super()._auth(
            username=username,
            password=password,
            pkey=pkey,
            key_filenames=key_filenames,
            allow_agent=allow_agent,
            look_for_keys=look_for_keys,
            gss_auth=gss_auth,
            gss_kex=gss_kex,
            gss_deleg_creds=gss_deleg_creds,
            gss_host=gss_host,
            passphrase=passphrase,
        )


paramiko.client.SSHClient = SSHClient



import time

import pandas as pd
from fabric.connection import Connection
from fabric.tasks import task

from xefab.utils import console, df_to_table


def parse_squeue_output(squeue_output):
    """Parse the output of the squeue command."""

    squeue_output = squeue_output.split("\n")
    header, rows = squeue_output[0], squeue_output[1:]
    header_fields = header.split()
    squeue_data = []
    for row in rows:
        row_data = {}
        fields = row.split()
        for name, field in zip(header_fields, fields):
            row_data[name] = field
        squeue_data.append(row_data)
    return pd.DataFrame(squeue_data, columns=header_fields).dropna(how="all")


@task(aliases=["job-queue"])
def squeue(
    c: Connection, user: str = "me", partition: str = None, out: str = ""
) -> pd.DataFrame:
    """Get the job-queue status."""

    command = 'squeue --format="%.18i %.9P %.30j %.8u %.8T %.10M %.9l %.6D %R"'

    if user in ["*", "all"]:
        pass
    elif user in ["me", "self", ""]:
        command += f" -u {c.user}"
    else:
        command += f" -u {user}"

    if partition:
        command += f" -p {partition}"

    with console.status(f"Running {command} on {c.host}..."):
        r = c.run(command, hide=True, warn=True)
        squeue_output = r.stdout
    if r.failed:
        console.print("Remote execution of squeue on {c.host} failed. stderr:")
        console.print(r.stderr)
        exit(r.return_code)

    df = parse_squeue_output(squeue_output)

    if out:
        with console.status(f"Saving squeue output to {out}..."):
            df.to_csv(out, index=False)
            time.sleep(0.5)
        console.print(f"Output written to {out}")
    else:
        table = df_to_table(df)
        if len(df) > 10:
            with console.pager():
                console.print(table)
        else:
            console.print(table)



import json
import random
import time
from io import BytesIO, StringIO
from random import choices
from string import ascii_lowercase

from fabric.tasks import task
from rich.progress import SpinnerColumn, TextColumn

from xefab.utils import ProgressContext, console, get_open_port

from .squeue import parse_squeue_output
from .utils import print_splash

JOB_NAME = "xefab-jupyter"


JOB_HEADER = """#!/bin/bash
#SBATCH --job-name={job_name}
#SBATCH --output={log_fn}
#SBATCH --error={log_fn}
#SBATCH --account=pi-lgrandi
#SBATCH --ntasks=1
#SBATCH --cpus-per-task={n_cpu}
#SBATCH --mem-per-cpu={mem_per_cpu}
#SBATCH --time={max_hours}:00:00
{extra_header}

export NUMEXPR_MAX_THREADS={n_cpu}
echo Starting jupyter job

"""

GPU_HEADER = """\
#SBATCH --partition=gpu2
#SBATCH --gres=gpu:1

module load cuda/10.1
"""

CPU_HEADER = """\
#SBATCH --qos {qos}
#SBATCH --partition {partition}
{reservation}
"""


# This is only if the user is NOT starting the singularity container
# (for singularity, starting jupyter is done in _xentenv_inner)
START_JUPYTER = """
echo $PYTHONPATH
jupyter {jupyter} --no-browser --port={port} --ip=0.0.0.0 --notebook-dir {notebook_dir} 2>&1
"""


START_NOTEBOOK_SH = """
#!/bin/bash

echo "Using singularity image: {CONTAINER}"

SINGULARITY_CACHEDIR=/scratch/midway2/{USER}/singularity_cache

module load singularity

singularity exec {BIND_STR} {CONTAINER} jupyter {JUPYTER_TYPE} --no-browser --port={PORT} --ip=0.0.0.0 --notebook-dir {NOTEBOOK_DIR}

"""

START_JUPYTER_SINGULARITY = """

SINGULARITY_CACHEDIR=/scratch/midway2/{user}/singularity_cache

module load singularity

singularity exec {bind_str} {container} jupyter {jupyter} --no-browser --port={port} --ip=0.0.0.0 --notebook-dir {notebook_dir}
"""


@task(
    pre=[print_splash],
    help={
        "env": "Environment to run on",
        "partition": "Partition to run on (xenon1t or dali)",
        "bypass_reservation": "Dont attempt to use the xenon notebook reservation",
        "tag": "Tag of the container to use",
        "binds": "Directories to bind to the container",
        "node": "Node to run on",
        "timeout": "Timeout for the job to start",
        "cpu": "Number of CPUs to request",
        "ram": "Amount of RAM to allocate (in MB)",
        "gpu": "Use a GPU",
        "jupyter": "Type of jupyter server to start (lab or notebook)",
        "local_cutax": "Use user installed cutax (from ~/.local)",
        "notebook_dir": "Directory to start the notebook in",
        "max_hours": "Maximum number of hours to run for",
        "force_new": "Force a new job to be started",
        "local_port": "Local port to attempt to forward to (if free)",
        "remote_port": "Port to use for jupyter server to on the worker node",
        "detached": "Run the job and exit, dont perform cleanup tasks.",
        "no_browser": "Dont open the browser automatically when done",
        "image_dir": "Directory to look for singularity images",
        "debug": "Print debug information",
    },
)
def start_jupyter(
    c,
    env: str = "singularity",
    partition: str = None,
    bypass_reservation: bool = False,
    tag: str = "development",
    binds: str = None,
    node: str = None,
    timeout: int = 120,
    cpu: int = 2,
    ram: int = 8000,
    gpu: bool = False,
    jupyter: str = "lab",
    local_cutax: bool = False,
    notebook_dir: str = None,
    max_hours: int = 4,
    force_new: bool = False,
    local_port: int = 8888,
    remote_port: int = None,
    detached: bool = False,
    no_browser: bool = False,
    image_dir: str = None,
    debug: bool = False,
):
    """Start a jupyter analysis notebook on the remote \
host and forward to local port via ssh-tunnel."""

    REMOTE_HOME = f"/home/{c.user}"

    unique_id = "".join(choices(ascii_lowercase, k=4))
    date = time.strftime("%Y%m%d")

    job_name = f"jupyter_{unique_id}_{date}"
    job_folder = f"{REMOTE_HOME}/xefab_jobs/{job_name}"

    server_details_path = f"{job_folder}/server_details.json"

    if image_dir is None:
        image_dir = "/project2/lgrandi/xenonnt/singularity-images"

    if notebook_dir is None:
        notebook_dir = REMOTE_HOME

    if remote_port is None:
        remote_port = random.randrange(15000, 20000)

    if partition is None:
        partition = "dali" if c.original_host == "dali" else "xenon1t"

    # bind directories inside the container
    if binds is None:
        binds = f"/project2,/scratch/midway2/{c.user},/dali"
    if isinstance(binds, str):
        binds = [bind.strip() for bind in binds.split(",")]
    bind_str = " ".join([f"--bind {bind}" for bind in binds])

    console.print(f"Using partition {partition}", style="info")

    local_port = get_open_port(start=local_port)
    env_vars = {}

    if local_cutax:
        console.warn(
            "Container cutax being overwritten by user installed package.",
            style="warning",
        )
        env_vars["INSTALL_CUTAX"] = "0"

    with ProgressContext() as progress:
        with progress.enter_task("Checking connection and destination folder"):
            if not c.run(f"test -d {job_folder}", warn=True).ok:
                progress.console.print(f"Creating {job_folder} on {c.host}")
                c.run("mkdir -p " + job_folder)

        if env == "singularity":
            with progress.enter_task("Building singularity runner script"):
                s_container = f"{image_dir}/xenonnt-{tag}.simg"
                starter_path = f"{job_folder}/start_notebook.sh"
                starter_script = START_NOTEBOOK_SH.format(
                    CONTAINER=s_container,
                    JUPYTER_TYPE=jupyter,
                    NOTEBOOK_DIR=notebook_dir,
                    PORT=remote_port,
                    BIND_STR=bind_str,
                    USER=c.user,
                )

            # Add the singularity runner script to the batch job
            # batch_job = JOB_HEADER + f"{starter_path} "
            batch_job = JOB_HEADER + START_JUPYTER_SINGULARITY.format(
                user=c.user,
                bind_str=bind_str,
                container=s_container,
                jupyter=jupyter,
                notebook_dir=notebook_dir,
                port=remote_port,
            )

            with progress.enter_task(
                "Uploading singularity runner script",
                finished_description="Singularity runner script uploaded",
            ) as task:
                starter_script_fd = StringIO(starter_script)
                c.put(starter_script_fd, remote=starter_path)
                progress.update(task, description="Making script executable")
                c.run(f"chmod +x {starter_path}")

        elif env == "cvmfs":
            batch_job = (
                JOB_HEADER
                + "source /cvmfs/xenon.opensciencegrid.org/releases/nT/%s/setup.sh"
                % (tag)
                + START_JUPYTER.format(
                    jupyter=jupyter, notebook_dir=notebook_dir, port=remote_port
                )
            )
            console.print(
                "Using conda from cvmfs (%s) instead of singularity container." % (tag),
                style="info",
            )

        elif env == "backup":
            if tag != "development":
                raise ValueError(
                    "I'm going to give you the latest container, you cannot choose a version!"
                )
            batch_job = (
                JOB_HEADER
                + "source /dali/lgrandi/strax/miniconda3/bin/activate strax"
                + START_JUPYTER.format(
                    jupyter=jupyter, notebook_dir=notebook_dir, port=remote_port
                )
            )
            console.print(
                "Using conda from cvmfs (%s) instead of singularity container." % (tag)
            )

        if partition == "kicp":
            qos = "xenon1t-kicp"
        else:
            qos = partition

        # FIXME: check if a job is already running.

        _want_to_make_reservation = partition == "xenon1t" and (not bypass_reservation)
        if ram > 16000 and _want_to_make_reservation:
            console.print(
                "You asked for more than 16 GB total memory you cannot use the notebook "
                "reservation queue for this job! We will bypass the reservation."
            )

        if cpu >= 8 and _want_to_make_reservation:
            console.print(
                "You asked for more than 7 CPUs you cannot use the notebook reservation "
                "queue for this job! We will bypass the reservation."
            )
        use_reservation = (
            (not force_new) and _want_to_make_reservation and cpu < 8 and ram <= 16000
        )

        if use_reservation:
            with progress.enter_task(
                "Notebook reservation requested. Checking availability",
                finished_description="using Notebook reservation.",
                exception_description="Notebook reservation does not exist, submitting a regular job.",
                warn=True,
                hide=not debug,
            ):
                result = c.run("scontrol show reservations", hide=True, warn=True)
                if (
                    result.failed
                    or "ReservationName=xenon_notebook" not in result.stdout
                ):
                    use_reservation = False
                    raise RuntimeError("Notebook reservation does not exist.")

        # with progress.enter_task("Checking for existing jobs"):
        #     result = c.run(f"squeue -u {c.user} -n straxlab", hide=True, warn=True)
        #     df = parse_squeue_output(result.stdout)
        job_fn = "/".join([job_folder, "notebook.sbatch"])
        log_fn = "/".join([job_folder, "notebook.log"])

        extra_header = (
            GPU_HEADER
            if gpu
            else CPU_HEADER.format(
                partition=partition,
                qos=qos,
                reservation=(
                    "#SBATCH --reservation=xenon_notebook" if use_reservation else ""
                ),
            )
        )

        if node:
            extra_header += "\n#SBATCH --nodelist={node}".format(node=node)
        if max_hours is None:
            max_hours = 2 if gpu else 8
        else:
            max_hours = int(max_hours)

        batch_job = batch_job.format(
            job_name=job_name,
            log_fn=log_fn,
            max_hours=max_hours,
            extra_header=extra_header,
            n_cpu=cpu,
            mem_per_cpu=int(ram / cpu),
        )

        with progress.enter_task("Reseting log file"):
            c.put(StringIO(""), remote=log_fn)

        with progress.enter_task("Copying batch job to remote host"):
            c.put(StringIO(batch_job), remote=job_fn)

        with progress.enter_task("Setting permissions on batch job"):
            c.run("chmod +x " + job_fn)

        with progress.enter_task("Submitting batch job"):
            result = c.run("sbatch " + job_fn, env=env_vars, hide=True, warn=True)

        if result.failed:
            raise RuntimeError("Could not submit batch job. Error: " + result.stderr)

        job_id = int(result.stdout.split()[-1])

        console.print("Submitted job with ID: %d" % job_id)

        with progress.enter_task(
            "Waiting for your job to start",
            finished_description="Job started.",
            exception_description="Job did not start.",
        ):
            for _ in range(timeout):
                result = c.run("squeue -j %d" % job_id, hide=True, warn=True)
                df = parse_squeue_output(result.stdout)
                if len(df) and df["ST"].iloc[0] == "R":
                    break
                elif len(df) and df["ST"].iloc[0] == "C":
                    raise RuntimeError("Job was cancelled.")
                time.sleep(1)
            else:
                raise RuntimeError("Timeout reached while waiting for job to start.")

        with progress.enter_task(
            "Waiting for jupyter server to start",
            finished_description="Jupyter server started.",
            exception_description="Jupyter server did not start.",
        ):
            url = None
            for _ in range(timeout):
                time.sleep(1)
                log_content = BytesIO()
                result = c.get(log_fn, local=log_content)
                log_content.seek(0)
                lines = [line.decode() for line in log_content.readlines()]

                for line in lines:
                    if "http://" in line and f":{remote_port}" in line:
                        url = line.split()[-1]
                        break
                else:
                    result = c.run("squeue -j %d" % job_id, hide=True, warn=True)
                    df = parse_squeue_output(result.stdout)
                    if not len(df):
                        raise RuntimeError(
                            "Job has exited."
                            "This can be because it was canceled or due to an internal error."
                        )
                    continue
                break
            else:
                raise RuntimeError(
                    "Timeout reached while waiting for jupyter to start."
                )

            console.print("\nJupyter started succesfully.")
            console.print(f"Remote URL: \n{url}\n")
            remote_host, remote_port = url.split("/")[2].split(":")
            if "token" in url:
                token = url.split("?")[1].split("=")[1]
                local_url = f"http://localhost:{local_port}?token={token}"
            else:
                token = ""
                local_url = f"http://localhost:{local_port}"

        # Can never be too careful, make sure port numbers are integers
        local_port = int(local_port)
        remote_port = int(remote_port)

        server_details = {
            "job_id": job_id,
            "job_name": job_name,
            "remote_host": remote_host,
            "remote_port": remote_port,
            "token": token,
        }

        # Persist server details
        with progress.enter_task(
            "Writing server details to file",
            finished_description="Server details saved",
            warn=True,
        ):
            details_fd = StringIO(json.dumps(server_details, indent=4))
            c.put(details_fd, remote=server_details_path)

        # Handle port forwarding
        msg = f"Forwarding remote address {remote_host}:{remote_port} to local port {local_port}"
        if not detached:
            extra_msg = (
                f"\nYou can access the notebook at \n{local_url}\n"
                "   Press ENTER or CTRL-C to deactivate and cancel job."
            )
            with progress.enter_task(
                msg + extra_msg,
                exception_description="Exception raised while forwarding port",
                warn=True,
            ) as task:
                with c.forward_local(
                    local_port, remote_port=remote_port, remote_host=remote_host
                ):
                    time.sleep(3)

                    if not no_browser:
                        result = c.local(
                            f"python -m webbrowser -t {local_url}", hide=True, warn=True
                        )
                    try:
                        r = progress.console.input()
                        progress.update(
                            task, description="Deactivating port forwarding"
                        )
                        time.sleep(2)
                    except KeyboardInterrupt:
                        console.print("Keyboard interrupt received.")
                    except Exception as e:
                        if debug:
                            console.print(f"Exception raised: {e}")
                        else:
                            console.print(f"Exception raised.")

            with progress.enter_task(
                "Canceling job",
                finished_description="Job canceled",
                exception_description=f"Could not cancel job. Please cancel it manually. Job ID: {job_id}",
                warn=True,
            ):
                c.run(f"scancel {job_id}", hide=True)

            if not debug:
                with progress.enter_task(
                    "Cleaning up job files",
                    finished_description="Job folder removed",
                    exception_description=f"Could not remove job folder. Please remove it manually. Path: {job_folder}",
                ):
                    for _ in range(3):
                        time.sleep(2)
                        result = c.run(f"rm -rf {job_folder}", hide=True, warn=True)
                        if result.ok:
                            break
                    else:
                        raise

        else:
            # Detached mode - just forward the port and exit
            with progress.enter_task(msg, warn=True) as task:
                result = c.local(
                    f"ssh -fN -L {local_port}:{remote_host}:{remote_port} {c.user}@{c.host} &",
                    disown=True,
                    hide=False,
                    warn=True,
                )
                time.sleep(3)
                console.print(f"You can access the notebook at \n{local_url}\n")
                if result.ok and not no_browser:
                    c.local(
                        f"python -m webbrowser -t {local_url}", hide=True, warn=True
                    )

    console.print("Goodbye!")
